/*
 * main.c: Datamodel daemon
 *
 * Copyright (C) 2023 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: Iryna Antsyferova <iryna.antsyferova@iopsys.eu>
 *
 * See LICENSE file for license related information.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <libubox/blobmsg.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <sys/prctl.h>
#include <sys/mman.h>

#include "set.h"
#include "get.h"
#include "add_delete.h"
#include "event.h"
#include "get_helper.h"
#include "cli.h"
#include "libdmtree/dmentry.h"

#define TELCOVOICEMGR_CONF_FILE "/etc/rdk/conf/telcovoice_manager_conf.json"

extern struct list_head global_memhead;
extern DM_MAP_OBJ tDynamicObj[];
extern DM_MAP_VENDOR tVendorExtension[];

static const char g_ubus_path[] = "voice.sip.data";

int rpc_handler_init()
{
	if (json_hal_server_init(TELCOVOICEMGR_CONF_FILE) != RETURN_OK)	{
		ERR("Failed to init HAL server");
		return -1;
	}

	if (json_hal_server_register_action_callback("setParameters",
		bbfdm_set_value) != RETURN_OK) {
		ERR("Failed to register setParameters callback");
		goto Error;
	}

	if (json_hal_server_register_action_callback("getParameters",
		bbfdm_get_value) != RETURN_OK) {
		ERR("Failed to register getParameters callback");
		goto Error;
	}

	if (json_hal_server_register_action_callback("deleteObject",
		bbfdm_del_object) != RETURN_OK) {
		ERR("Failed to register deleteObject callback");
		goto Error;
	}
	if (json_hal_server_register_action_callback("subscribeEvent",
		subs_event_handler) != RETURN_OK) {
		ERR("Failed to register subscribeEvent callback");
		goto Error;
	}

	if (json_hal_server_run() != RETURN_OK) {
		ERR("Failed to run HAL server");
		goto Error;
	}

	return 0;

Error:
	json_hal_server_terminate();
	return -1;
}

int rpc_handler_uninit()
{
	if (json_hal_server_terminate() != RETURN_OK) {
		ERR("Failed to terminate HAL server");
		return -1;
	}

	return 0;
}

static void usage(char *prog)
{
	fprintf(stderr, "Usage: %s [options]\n", prog);
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "    -c <command input>  Run cli command\n");
	fprintf(stderr, "    -h                 Displays this help\n");
	fprintf(stderr, "\n");
}

static void bbfdm_cleanup(void)
{
	bbf_global_clean(DEAMON_DM_ROOT_OBJ);
}

int bbfdm_load_deamon_config(void)
{
	DEAMON_DM_ROOT_OBJ = (DMOBJ *)dm_dynamic_calloc(&global_memhead, 2, sizeof(DMOBJ));
	if (DEAMON_DM_ROOT_OBJ == NULL) {
		fprintf(stderr, "ERROR: No Memory exists\n");
		return -1;
	}

	char *node_obj = dm_dynamic_strdup(&global_memhead, tDynamicObj[0].path);
	unsigned int len = strlen(node_obj);

	node_obj[len-1] = 0;
	DEAMON_DM_ROOT_OBJ[0].obj = node_obj; 
	DEAMON_DM_ROOT_OBJ[0].nextobj = tDynamicObj[0].root_obj;
	DEAMON_DM_ROOT_OBJ[0].leaf = tDynamicObj[0].root_leaf;
	DEAMON_DM_ROOT_OBJ[0].permission = &DMREAD;
	DEAMON_DM_ROOT_OBJ[0].bbfdm_type = BBFDM_BOTH;

	DEAMON_DM_VENDOR_EXTENSION[0] = tVendorExtension;
	//DEAMON_DM_VENDOR_EXTENSION[1] = tVendorExtensionOverwrite;
	//DEAMON_DM_VENDOR_EXTENSION_EXCLUDE = tVendorExtensionExclude;

	return 0;
}

static void program_exit(int ret)
{
	rpc_handler_uninit();
	bbfdm_cleanup();
	closelog();
	uloop_done();
	exit(ret);
}

static void sig_handler(__attribute__((unused)) int signum)
{
	program_exit(0);
}

static void signal_init()
{
	signal(SIGINT, sig_handler);
}

static void voice_event_cb(struct ubus_context *ctx __attribute__((unused)),
		struct ubus_event_handler *ev __attribute__((unused)),
		const char *type, struct blob_attr *msg)
{
	const struct blobmsg_policy policy[] = {
		[VOICE_CON_ENABLED] = { .name = "enabled", .type = BLOBMSG_TYPE_BOOL},
		[VOICE_CON_PROTOCOL] = { .name = "protocol", .type = BLOBMSG_TYPE_STRING},
		[VOICE_CON_PORT] = { .name = "port", .type = BLOBMSG_TYPE_INT32},
		[VOICE_CON_IP] = { .name = "ip", .type = BLOBMSG_TYPE_STRING},
	};
	struct blob_attr *tb[_VOICE_CON_MAX];

	if (type && strcmp(type, g_ubus_path) != 0)
		return;

	if (blobmsg_parse(policy, _VOICE_CON_MAX, tb, blob_data(msg), blob_len(msg))){
		ERR("Failed to parse blob");
		return;
	}

	send_fw_rule_event(blobmsg_get_bool(tb[VOICE_CON_ENABLED]),
		blobmsg_get_string(tb[VOICE_CON_PROTOCOL]),
		blobmsg_get_u32(tb[VOICE_CON_PORT]),
		blobmsg_get_string(tb[VOICE_CON_IP]), 1);
}

int main(int argc, char **argv)
{
	int err = 0, ch;
	struct ubus_context *ubus_ctx = NULL;
	struct ubus_event_handler add_event = {};

	signal_init();

	while ((ch = getopt(argc, argv, "h:c:")) != -1) {
		switch (ch) {
		case 'c':
			err = bbfdm_cli_exec_command(argc-optind+1, &argv[optind-1]);
			exit(err);
		case 'h':
			usage(argv[0]);
			exit(0);
		default:
			break;
		}
	}

	openlog(argv[0], LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	err = bbfdm_load_deamon_config();
	if (err) {
		ERR("Failed to configure DM");
		goto Exit;
	}

	err = rpc_handler_init();
	if (err) {
		ERR("Voice HAL Server Initialization Failed");
		goto Exit;
	}

	uloop_init();

	ubus_ctx = ubus_connect(NULL);
	if (!ubus_ctx) {
		ERR("Failed to connect to ubus");
		return -1;
	}

	ubus_add_uloop(ubus_ctx);
	add_event.cb = voice_event_cb;
	if (ubus_register_event_handler(ubus_ctx, &add_event, g_ubus_path))
		ERR("Failed to connect to ubus");

	INFO("Waiting on uloop....");
	uloop_run();

Exit:
	ubus_shutdown(ubus_ctx);
	program_exit(err);

	return err;
}
